from time import sleep

from efidgy import impl

from . import idd_or


__all__ = [
    idd_or,
    # 'Computation',
    # 'Issue',
    # 'IssueStats',
    # 'IssueType',
    # 'Issues',
    # 'JobMessage',
    'JobState',
    'Member',
    'Project',
    'ProjectState',
    # 'ProjectSummary',
    'ProjectType',
    'ProjectTypeCode',
    # 'Role',
    'SharedMode',
    'Solution',
    # 'SolutionSummary',
    'UnitSystem',
]


class UnitSystem:
    METRIC = 'metric'
    IMPERIAL = 'imperial'


class ProjectTypeCode:
    IDD_OR = 'idd_or'


class SharedMode:
    PRIVATE = 'private'
    READONLY = 'readonly'
    EDITABLE = 'editable'


class ProjectState:
    IDLE = 'idle'
    COMPUTATING = 'computating'
    IMPORTING = 'importing'


class JobState:
    PENDING = 'pending'
    WORKING = 'working'
    ERROR = 'error'
    SUCCESS = 'success'


class ICurrency(impl.model.Model):
    code = impl.fields.PrimaryKey()
    symbol = impl.fields.CharField()


class IProjectType(impl.model.Model):
    code = impl.fields.PrimaryKey()
    name = impl.fields.CharField()
    description = impl.fields.CharField()


class IMember(impl.model.Model):
    pk = impl.fields.PrimaryKey()
    email = impl.fields.CharField()
    first_name = impl.fields.CharField()
    last_name = impl.fields.CharField()
    role = impl.fields.CharField()


class IProject(impl.model.Model):
    pk = impl.fields.PrimaryKey()
    name = impl.fields.CharField()
    currency = impl.fields.ObjectField(model=ICurrency)
    project_type = impl.fields.ObjectField(model=IProjectType)
    shared_mode = impl.fields.CharField()
    owner = impl.fields.ObjectField(model=IMember)
    state = impl.fields.CharField()
    outdated = impl.fields.BooleanField()
    progress = impl.fields.FloatField()
    issue_stats = impl.fields.DictField()
    summary = impl.fields.DictField()

    def _get_computation_path(self):
        return '{}/{}/computation'.format(
            self.service._get_path(self._get_context()),
            self.pk,
        )

    def _log_messages(self, messages):
        logged_messages = getattr(self, '_logged_messages', 0)
        for i, message in enumerate(messages):
            if i < logged_messages:
                continue
            impl.module_logger.info(message['message'])
            logged_messages += 1
        self._logged_messages = logged_messages


class ISolution(impl.model.Model):
    pk = impl.fields.PrimaryKey()
    cost = impl.fields.FloatField()
    outdated = impl.fields.BooleanField()
    issue_stats = impl.fields.DictField()
    summary = impl.fields.DictField()


class Member(IMember):
    pass


class Currency(ICurrency):
    class service(impl.service.SyncViewMixin, impl.service.EfidgyService):
        path = '/refs/currencies'


class ProjectType(IProjectType):
    class service(impl.service.SyncViewMixin, impl.service.EfidgyService):
        path = '/refs/project_types'


class Project(IProject):
    currency = impl.fields.ObjectField(model=Currency)
    project_type = impl.fields.ObjectField(model=ProjectType)
    member = impl.fields.ObjectField(model=Member)

    class service(impl.service.SyncChangeMixin, impl.service.CustomerService):
        path = '/projects'

    def start_computation(self):
        c = impl.client.SyncClient(self.service._get_env())
        c.post(self._get_computation_path(), None)

    def stop_computation(self):
        c = impl.client.SyncClient(self.service._get_env())
        c.delete(self._get_computation_path())

    def wait_computation(self):
        while True:
            c = impl.client.SyncClient(self.service._get_env())
            response = c.get(self._get_computation_path())
            self._log_messages(response['messages'])
            if response['state'] not in [JobState.PENDING, JobState.WORKING]:
                break
            sleep(10)

    def computate(self):
        self.start_computation()
        self.wait_computation()


class Solution(ISolution):
    class service(impl.service.SyncViewMixin, impl.service.ProjectService):
        path = '/solutions'



class EfidgyException(Exception):
    pass


class VersionError(EfidgyException):
    def __init__(self, client_version, server_version):
        super().__init__(
            'Update client from version {} to version: {}'.format(
                client_version,
                server_version,
            ),
        )


class BadRequest(EfidgyException):
    pass


class ValidationError(BadRequest):
    pass


class AuthenticationFailed(EfidgyException):
    pass


class PermissionDeined(EfidgyException):
    pass


class NotFound(EfidgyException):
    pass


class MethodNotAllowed(EfidgyException):
    pass


class InternalServerError(EfidgyException):
    pass


class GeocodeError(EfidgyException):
    def __init__(self, address):
        super().__init__(address)
        self.address = address



from efidgy import impl


__all__ = [
    'Order',
    'OrderRoute',
    'OrderSchedule',
    'Path',
    'Point',
    'PointType',
    'Store',
    'Vehicle',
    'VehicleRoute',
    'VehicleSchedule',
]


class PointType:
    STORE = 'store'
    ORDER = 'order'


class IPoint(impl.model.Model):
    pk = impl.fields.PrimaryKey()
    address = impl.fields.CharField()
    enabled = impl.fields.BooleanField()
    lat = impl.fields.FloatField()
    lon = impl.fields.FloatField()
    point_type = impl.fields.CharField()


class IPath(impl.model.Model):
    pk = impl.fields.PrimaryKey()
    distance = impl.fields.FloatField()
    directions = impl.fields.CharField()


class ISchedule(impl.model.Model):
    pk = impl.fields.PrimaryKey()
    start_point = impl.fields.PolymorphObjectField(
        lookup_field='point_type',
        models={
            PointType.STORE: 'efidgy.models.idd_or.IStore',
            PointType.ORDER: 'efidgy.models.idd_or.IOrder',
        },
    )
    end_point = impl.fields.PolymorphObjectField(
        lookup_field='point_type',
        models={
            PointType.STORE: 'efidgy.models.idd_or.IStore',
            PointType.ORDER: 'efidgy.models.idd_or.IOrder',
        },
    )
    departure_time = impl.fields.TimeField()
    arrival_time = impl.fields.TimeField()
    path = impl.fields.ObjectField(model=IPath)


class IStore(IPoint):
    name = impl.fields.CharField()
    description = impl.fields.CharField()
    open_time = impl.fields.TimeField()
    close_time = impl.fields.TimeField()
    issues = impl.fields.DictField()


class IVehicle(impl.model.Model):
    pk = impl.fields.PrimaryKey()
    name = impl.fields.CharField()
    description = impl.fields.CharField()
    enabled = impl.fields.BooleanField()
    store = impl.fields.ObjectField(model=IStore)
    features = impl.fields.ListField(item=impl.fields.CharField())
    fuel_consumption = impl.fields.FloatField()
    fuel_price = impl.fields.FloatField()
    salary_per_distance = impl.fields.FloatField()
    salary_per_duration = impl.fields.FloatField()
    items_limit = impl.fields.IntegerField()
    volume_limit = impl.fields.FloatField()
    weight_limit = impl.fields.FloatField()
    start_time = impl.fields.TimeField()
    end_time = impl.fields.TimeField()
    duration_limit = impl.fields.DurationField()
    route = impl.fields.ObjectField(model='efidgy.models.idd_or.IVehicleRoute')
    issues = impl.fields.DictField()


class IVehicleRoute(impl.model.Model):
    schedule = impl.fields.ListField(
        item='efidgy.models.idd_or.IVehicleSchedule',
    )
    distance = impl.fields.FloatField()
    distance_salary = impl.fields.FloatField()
    start_time = impl.fields.TimeField()
    duration = impl.fields.DurationField()
    duration_salary = impl.fields.FloatField()
    fuel = impl.fields.FloatField()
    fuel_cost = impl.fields.FloatField()
    issues = impl.fields.DictField()


class IVehicleSchedule(ISchedule):
    orders = impl.fields.ListField(item='efidgy.models.idd_or.IOrder')
    issues = impl.fields.DictField()


class IOrder(IPoint):
    name = impl.fields.CharField()
    description = impl.fields.CharField()
    features = impl.fields.ListField(item=impl.fields.CharField())
    items = impl.fields.IntegerField()
    volume = impl.fields.FloatField()
    weight = impl.fields.FloatField()
    ready_time = impl.fields.TimeField()
    store = impl.fields.ObjectField(model=IStore)
    delivery_time_from = impl.fields.TimeField()
    delivery_time_to = impl.fields.TimeField()
    load_duration = impl.fields.DurationField()
    unload_duration = impl.fields.DurationField()
    route = impl.fields.ObjectField(model='efidgy.models.idd_or.IOrderRoute')
    issues = impl.fields.DictField()


class IOrderRoute(impl.model.Model):
    vehicle = impl.fields.ObjectField(model='efidgy.models.idd_or.IVehicle')
    schedule = impl.fields.ListField(
        item='efidgy.models.idd_or.IOrderSchedule',
    )
    issues = impl.fields.DictField()


class IOrderSchedule(ISchedule):
    issues = impl.fields.DictField()


class Point(IPoint):
    pass


class Path(IPath):
    pass


class Store(IStore):
    class service(impl.service.SyncChangeMixin, impl.service.ProjectService):
        path = '/stores'


class Vehicle(IVehicle):
    store = impl.fields.ObjectField(model=Store)
    route = impl.fields.ObjectField(model='efidgy.models.idd_or.VehicleRoute')

    class service(impl.service.SyncChangeMixin, impl.service.SolutionService):
        path = '/vehicles'


class VehicleRoute(IVehicleRoute):
    schedule = impl.fields.ListField(
        item='efidgy.models.idd_or.VehicleSchedule',
    )


class VehicleSchedule(IVehicleSchedule):
    start_point = impl.fields.PolymorphObjectField(
        lookup_field='point_type',
        models={
            PointType.STORE: 'efidgy.models.idd_or.Store',
            PointType.ORDER: 'efidgy.models.idd_or.Order',
        },
    )
    end_point = impl.fields.PolymorphObjectField(
        lookup_field='point_type',
        models={
            PointType.STORE: 'efidgy.models.idd_or.Store',
            PointType.ORDER: 'efidgy.models.idd_or.Order',
        },
    )
    path = impl.fields.ObjectField(model=Path)
    orders = impl.fields.ListField(item='efidgy.models.idd_or.Order')


class Order(IOrder):
    store = impl.fields.ObjectField(model=Store)
    route = impl.fields.ObjectField(model='efidgy.models.idd_or.OrderRoute')

    class service(impl.service.SyncChangeMixin, impl.service.SolutionService):
        path = '/orders'


class OrderRoute(IOrderRoute):
    vehicle = impl.fields.ObjectField(model='efidgy.models.idd_or.Vehicle')
    schedule = impl.fields.ListField(item='efidgy.models.idd_or.OrderSchedule')


class OrderSchedule(IOrderSchedule):
    start_point = impl.fields.PolymorphObjectField(
        lookup_field='point_type',
        models={
            PointType.STORE: Store,
            PointType.ORDER: Order,
        },
    )
    end_point = impl.fields.PolymorphObjectField(
        lookup_field='point_type',
        models={
            PointType.STORE: Store,
            PointType.ORDER: Order,
        },
    )
    path = impl.fields.ObjectField(model=Path)



import asyncio
import json
import re
import ssl
import urllib.parse
import urllib.request

import efidgy


class Client:
    _verified = False

    def __init__(self, env):
        self.env = env
        self._context = None
        if self.env.insecure:
            self._context = ssl._create_unverified_context()

    def _url(self, path):
        args = ['time_format=clock_24']
        if self.env.unit_system is not None:
            args.append('unit_system={}'.format(self.env.unit_system))
        o = urllib.parse.urlparse(path)
        path = o.path
        if path.endswith('/'):
            path = path[:-1]
        if o.query:
            args.append(o.query)
        assert self.env.code, (
            'No customer code defined.'
        )
        return 'https://{host}/api/{code}{path}/?{args}'.format(
            host=self.env.host,
            code=self.env.code,
            path=path,
            args='&'.join(args),
        )

    def _auth(self, request):
        if self.env.token:
            request.add_header(
                'Authorization',
                'Token {}'.format(self.env.token),
            )

    def _parse_version(self, version):
        m = re.match(r'(\d+)(?:\.(\d+))?(?:\.(\d+))?', version)

        major = int(m[1])

        minor = m[2]
        if minor is not None:
            minor = int(minor)

        patch = m[3]
        if patch is not None:
            patch = int(patch)

        return major, minor, patch

    def _match_versions(self, client_version, server_version):
        print('match', client_version, server_version)
        (
            client_major,
            client_minor,
            client_patch,
        ) = self._parse_version(client_version)
        (
            server_major,
            server_minor,
            server_patch,
        ) = self._parse_version(server_version)
        if client_major != server_major or client_minor != server_minor:
            raise efidgy.exceptions.VersionError(
                client_version,
                server_version,
            )

    def _check_api_version(self):
        if self._verified:
            return

        url = 'https://{}/efidgy_version.json'.format(self.env.host)
        try:
            with urllib.request.urlopen(url, context=self._context) as f:
                data = json.load(f)
        except (urllib.error.HTTPError, json.decoder.JSONDecodeError):
            data = None

        server_version = data.get('version') if data else None
        assert server_version, (
            'Unable to fetch server version.'
        )
        if server_version == 'dev':
            return
        if efidgy.__version__ == 'dev':
            return
        m = re.match(r'([^-]+)(?:-.+)?', efidgy.__version__)
        client_version = m[1] if m else efidgy.__version__
        self._match_versions(client_version, server_version)

        type(self)._verified = True

    def _urlopen(self, url, data, method):
        self._check_api_version()

        data = json.dumps(data).encode('utf-8') if data is not None else None

        req = urllib.request.Request(url, data=data, method=method)
        self._auth(req)
        if data:
            req.add_header('Content-Type', 'application/json')

        try:
            with urllib.request.urlopen(req, context=self._context) as f:
                return json.load(f)
        except urllib.error.HTTPError as e:
            try:
                data = json.load(e)
            except json.decoder.JSONDecodeError:
                data = {}
            if e.code == 400:
                detail = data.get('detail')
                if detail is not None:
                    raise efidgy.exceptions.BadRequest(detail) from None
                raise efidgy.exceptions.ValidationError(data) from None
            if e.code == 401:
                detail = data.get(
                    'detail',
                    'Authentication failed.',
                )
                raise efidgy.exceptions.AuthenticationFailed(detail) from None
            if e.code == 403:
                detail = data.get(
                    'detail',
                    'Permission denied.',
                )
                raise efidgy.exceptions.PermissionDenied(detail) from None
            if e.code == 404:
                detail = data.get('detail', 'Not found.')
                raise efidgy.exceptions.NotFound(detail) from None
            if e.code == 405:
                detail = data.get('detail', 'Method not allowed.')
                raise efidgy.exceptions.MethodNotAllowed(detail) from None
            if e.code >= 500 and e.code < 600:
                raise efidgy.exceptions.InternalServerError() from None
            raise
        except json.decoder.JSONDecodeError:
            return None


class SyncClient(Client):
    def get(self, path):
        return self._urlopen(self._url(path), None, 'GET')

    def post(self, path, data):
        return self._urlopen(self._url(path), data, 'POST')

    def put(self, path, data):
        return self._urlopen(self._url(path), data, 'PUT')

    def delete(self, path):
        return self._urlopen(self._url(path), None, 'DELETE')


class AsyncClient(Client):
    async def _async_urlopen(self, *args):
        return await asyncio.get_event_loop().run_in_executor(
            None,
            self._urlopen,
            *args,
        )

    async def get(self, path):
        return await self._async_urlopen(self._url(path), None, 'GET')

    async def post(self, path, data):
        return await self._async_urlopen(self._url(path), data, 'POST')

    async def put(self, path, data):
        return await self._async_urlopen(self._url(path), data, 'PUT')

    async def delete(self, path):
        return await self._async_urlopen(self._url(path), None, 'DELETE')



import efidgy

from . import fields


class Meta:
    def __init__(self, fields, primary_key=None):
        self.fields = fields
        self.primary_key = primary_key


class ModelMeta(type):
    @classmethod
    def _iter_fields(cls, obj):
        for base in obj.__bases__:
            for field_name, field in cls._iter_fields(base):
                yield field_name, field

            for field_name, field in vars(obj).items():
                if not isinstance(field, fields.Field):
                    continue
                yield field_name, field

    @classmethod
    def _repr(cls):
        def repr_impl(self):
            fields = []
            for field in self._meta.fields:
                value = getattr(self, field.name, None)
                if isinstance(value, str):
                    value = '"{}"'.format(value)
                fields.append('{}={}'.format(field.name, value))
            return '<{} {}>'.format(self.__class__.__name__, ' '.join(fields))
        return repr_impl

    def __new__(cls, name, bases, attrs):
        Model = super().__new__(cls, name, bases, attrs)

        Model.__repr__ = cls._repr()

        fields = []
        primary_key = None
        for field_name, field in cls._iter_fields(Model):
            field.name = field_name
            fields.append(field)
            if (
                field.primary_key
            ):
                assert primary_key is None, (
                    'Multiple primary keys are not allowed.'
                )
                primary_key = field

        Model._meta = Meta(fields, primary_key=primary_key)

        service = getattr(Model, 'service', None)
        if service is not None:
            service._bind(Model)

        return Model


class Model(metaclass=ModelMeta):
    _env = None

    @classmethod
    def _decode(cls, data, **kwargs):
        kw = {**kwargs}
        for field in cls._meta.fields:
            kw[field.name] = field._decode(data.get(field.name), **kwargs)
        return cls(**kw)

    def _encode(self):
        ret = {}
        for field in self._meta.fields:
            value = field._encode(getattr(self, field.name))
            if value is not None:
                ret[field.name] = value
        return ret

    def __eq__(self, other):
        if type(self) != type(other):
            return False
        if (
            self._meta.primary_key is None
            or other._meta.primary_key is None
        ):
            return self == other
        self_pk = getattr(self, self._meta.primary_key.name)
        other_pk = getattr(other, other._meta.primary_key.name)
        return self_pk == other_pk

    def _get_context(self):
        return {}

    def __init__(self, **kwargs):
        service = getattr(self, 'service', None)
        if service is not None:
            self.service = self.service()

        for field in self._meta.fields:
            setattr(self, field.name, kwargs.get(field.name))



import unittest

import datetime

import asyncio

import efidgy
from efidgy import models
from efidgy import tools
from efidgy import exceptions
from efidgy.asyncapi import models as amodels
from efidgy.asyncapi import tools as atools

import logging


def async_test(coro):
    def wrapper(*args, **kwargs):
        loop = asyncio.new_event_loop()
        try:
            return loop.run_until_complete(coro(*args, **kwargs))
        finally:
            loop.close()
    return wrapper


class TestImpl(unittest.TestCase):
    PROJECT_NAME = 'Test Project'

    def setUp(self):
        efidgy.env = efidgy.env.override(
            unit_system=efidgy.models.UnitSystem.IMPERIAL,
        )
        projects = models.Project.service.filter(name=self.PROJECT_NAME)
        for project in projects:
            project.delete()

    def test_client_errors(self):
        with self.assertRaises(AssertionError):
            models.ProjectType.service.get(pk='XXX')
        with self.assertRaises(AssertionError):
            models.Project.service.get(foo='XXX')

    def test_authentication(self):
        env = efidgy.env.override(token='XXX')
        with self.assertRaises(exceptions.AuthenticationFailed):
            models.Project.service.use(env).get(pk='XXX')

    def test_not_found(self):
        with self.assertRaises(exceptions.NotFound):
            models.Project.service.get(pk='XXX')

    def test_validation(self):
        with self.assertRaises(exceptions.ValidationError):
            models.Project.service.create(
                name=self.PROJECT_NAME,
                currency='USD',
                project_type='XXX',
                shared_mode=models.SharedMode.PRIVATE,
            )

    def test_use(self):
        models.ProjectType.service.all()
        models.ProjectType.service.use(efidgy.env.override(code='XXX')).all()

    def test_filter(self):
        project = models.Project.service.create(
            name=self.PROJECT_NAME,
            currency='USD',
            project_type=models.ProjectTypeCode.IDD_OR,
            shared_mode=models.SharedMode.PRIVATE,
        )
        all_projects = models.Project.service.filter(name=self.PROJECT_NAME)
        first_project = models.Project.service.first(name=self.PROJECT_NAME)

        self.assertEqual(len(all_projects), 1)
        self.assertEqual(first_project, project)

        store = models.idd_or.Store.service.create(
            project=project,
            address='Address',
            name='Delivery Inc.',
            lat=0,
            lon=0,
        )

        stores = models.idd_or.Store.service.all(project=project)
        self.assertEqual(len(stores), 1)

        stores = models.idd_or.Store.service.filter(
            project=project,
            name='Delivery Inc.',
        )
        self.assertEqual(len(stores), 1)

        first_store = models.idd_or.Store.service.first(
            project=project,
            name='Delivery Inc.',
        )
        self.assertEqual(first_store, store)

    @async_test
    async def test_avalidation(self):
        with self.assertRaises(exceptions.ValidationError):
            await amodels.Project.service.create(
                name=self.PROJECT_NAME,
                currency='USD',
                project_type='XXX',
                shared_mode=amodels.SharedMode.PRIVATE,
            )

    def test_sync(self):
        project = models.Project.service.create(
            name=self.PROJECT_NAME,
            currency='USD',
            project_type=models.ProjectTypeCode.IDD_OR,
            shared_mode=models.SharedMode.PRIVATE,
        )
        project.delete()

    @async_test
    async def test_async(self):
        project = await amodels.Project.service.create(
            name=self.PROJECT_NAME,
            currency='USD',
            project_type=amodels.ProjectTypeCode.IDD_OR,
            shared_mode=amodels.SharedMode.PRIVATE,
        )
        await project.delete()


class TestModels(unittest.TestCase):
    PROJECT_NAME = 'Test Project'

    def setUp(self):
        # logging.basicConfig(
        #     format='%(asctime)s %(name)s %(message)s',
        #     level=logging.DEBUG,
        # )

        for project in models.Project.service.all():
            if project.name == self.PROJECT_NAME:
                project.delete()

        self.stores = [
            {
                'address': '6133 Broadway Terr., Oakland, CA 94618, USA',
                'name': 'Delivery Inc.',
                'open_time': datetime.time(8, 0),
                'close_time': datetime.time(18, 0),
            },
        ]

        self.vehicles = [
            {
                'store': 'Delivery Inc.',
                'name': 'Gary Bailey',
                'fuel_consumption': 11.76,
                'fuel_price': 3.25,
                'salary_per_duration': 21,
                'duration_limit': datetime.timedelta(hours=9),
            },
        ]

        self.orders = [
            {
                'store': 'Delivery Inc.',
                'name': '#00001',
                'address': '1 Downey Pl, Oakland, CA 94610, USA',
                'ready_time': datetime.time(8, 0),
                'delivery_time_from': datetime.time(12, 0),
                'delivery_time_to': datetime.time(16, 0),
                'load_duration': datetime.timedelta(minutes=1),
                'unload_duration': datetime.timedelta(minutes=5),
                'items': 1,
                'volume': 3.53,
                'weight': 22.05,
            },
        ]

    def _repr_point(self, point):
        return point.name

    def _print_vehicle(self, vehicle):
        print(vehicle.name)
        if vehicle.route is None:
            return
        prev_schedule = None
        for schedule in vehicle.route.schedule:
            print('{at}\t{arr}\t{dep}'.format(
                at=self._repr_point(schedule.start_point),
                arr=prev_schedule.arrival_time if prev_schedule else '',
                dep=schedule.departure_time,
            ))
            prev_schedule = schedule
        if prev_schedule:
            print('{at}\t{arr}\t{dep}'.format(
                at=self._repr_point(prev_schedule.end_point),
                arr=prev_schedule.arrival_time,
                dep='',
            ))

    def _print_order(self, order):
        print(order.name)
        if order.route is None:
            return
        prev_schedule = None
        for schedule in order.route.schedule:
            print('{at}\t{arr}\t{dep}'.format(
                at=self._repr_point(schedule.start_point),
                arr=prev_schedule.arrival_time if prev_schedule else None,
                dep=schedule.departure_time,
            ))
            prev_schedule = schedule
        if prev_schedule:
            print('{at}\t{arr}\t{dep}'.format(
                at=self._repr_point(prev_schedule.end_point),
                arr=prev_schedule.arrival_time,
                dep=None,
            ))

    def test_solve(self):
        project = models.Project.service.create(
            name=self.PROJECT_NAME,
            currency='USD',
            project_type=models.ProjectTypeCode.IDD_OR,
            shared_mode=models.SharedMode.PRIVATE,
        )

        stores = {}
        for data in self.stores:
            lat, lon = tools.geocode(data['address'])
            store = models.idd_or.Store.service.create(
                project=project,
                lat=lat,
                lon=lon,
                **data,
            )
            stores[store.name] = store

        vehicles = {}
        for data in self.vehicles:
            data['store'] = stores[data['store']]
            vehicle = models.idd_or.Vehicle.service.create(
                project=project,
                **data,
            )
            vehicles[vehicle.name] = vehicle

        orders = {}
        for data in self.orders:
            lat, lon = tools.geocode(data['address'])
            data['store'] = stores[data['store']]
            order = models.idd_or.Order.service.create(
                project=project,
                lat=lat,
                lon=lon,
                **data,
            )
            orders[order.name] = order

        project.computate()

        solutions = models.Solution.service.all(
            project=project,
        )
        self.assertTrue(len(solutions) > 0)
        solution = solutions[0]
        print('{cost:.2f}{currency}'.format(
            cost=solution.cost,
            currency=project.currency.symbol
        ))

        vehicles = models.idd_or.Vehicle.service.all(
            project=project,
            solution=solution,
        )
        for vehicle in vehicles:
            self._print_vehicle(vehicle)

        orders = models.idd_or.Order.service.all(
            project=project,
            solution=solution,
        )
        for order in orders:
            self._print_order(order)

    @async_test
    async def test_asolve(self):
        project = await amodels.Project.service.create(
            name=self.PROJECT_NAME,
            currency='USD',
            project_type=amodels.ProjectTypeCode.IDD_OR,
            shared_mode=amodels.SharedMode.PRIVATE,
        )

        stores = {}
        for data in self.stores:
            lat, lon = await atools.geocode(data['address'])
            store = await amodels.idd_or.Store.service.create(
                project=project,
                lat=lat,
                lon=lon,
                **data,
            )
            stores[store.name] = store

        vehicles = {}
        for data in self.vehicles:
            data['store'] = stores[data['store']]
            vehicle = await amodels.idd_or.Vehicle.service.create(
                project=project,
                **data,
            )
            vehicles[vehicle.name] = vehicle

        orders = {}
        for data in self.orders:
            lat, lon = await atools.geocode(data['address'])
            data['store'] = stores[data['store']]
            order = await amodels.idd_or.Order.service.create(
                project=project,
                lat=lat,
                lon=lon,
                **data,
            )
            orders[order.name] = order

        await project.computate()

        solutions = await amodels.Solution.service.all(
            project=project,
        )
        self.assertTrue(len(solutions) > 0)
        solution = solutions[0]
        print('{cost:.2f}{currency}'.format(
            cost=solution.cost,
            currency=project.currency.symbol
        ))

        vehicles = await amodels.idd_or.Vehicle.service.all(
            project=project,
            solution=solution,
        )
        for vehicle in vehicles:
            self._print_vehicle(vehicle)

        orders = await amodels.idd_or.Order.service.all(
            project=project,
            solution=solution,
        )
        for order in orders:
            self._print_order(order)
